from abc import abstractmethod, ABC
from collections import defaultdict
from typing import Tuple, List, Optional, Union
from uuid import uuid4

import numpy as np

from sampo.scheduler.base import Scheduler
from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.graph import WorkGraph
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import ScheduleWorkDict, Schedule
from sampo.schemas.time import Time
from sampo.utilities.visualization.resources import get_workers_intervals, SPLITTER


class ResourceOptimizer(ABC):
    @abstractmethod
    def optimize(self, wg: WorkGraph, deadline: Time) -> Tuple[Contractor, Time]:
        ...


# TODO: remove ScheduleWorkDict option
def max_time_schedule(schedule: Union[ScheduleWorkDict, Schedule]):
    return max(work["finish"] for work in schedule.values()) \
            if isinstance(schedule, ScheduleWorkDict) \
            else max(swork.start_end_time[1] for swork in schedule.works)


def get_schedule_with_time(scheduler: Scheduler, wg: WorkGraph,
                           agent_counts: np.array, agent_names: List[str]) -> Tuple[Schedule, Time]:
    scheduled_works = scheduler.schedule(wg, agents_to_contractors(agent_counts, agent_names))
    max_time = max_time_schedule(scheduled_works)
    return scheduled_works, max_time


def agents_to_contractors(agent_counts: np.array, agent_names: List[str],
                          contractor_id: Optional[str or None] = None,
                          contractor_name: Optional[str] = ""):
    contractor_id = contractor_id or uuid4()
    # TODO Remove...
    # TODO: remove try
    try:
        workers = [Worker(str(uuid4()), name, count) for name, count in zip(agent_names, agent_counts)]
    except Exception:
        print('обосрамс...')
    workers_dict = {(w.name, w.productivity_class): w for w in workers}
    return [Contractor(contractor_id, contractor_name, agent_names, [], workers_dict, {})]


def is_resources_good(wg: WorkGraph,
                      agent_counts: np.array, agent_names: List[str],
                      scheduler: Scheduler, deadline: Time) -> bool:
    try:
        _, schedule_time = get_schedule_with_time(scheduler, wg, agent_counts, agent_names)
        return schedule_time <= deadline
    # TODO: remove
    except AssertionError:
        return False
    except Exception as ex:
        print(ex)
        return False


def find_min_agents(wg: WorkGraph, max_workers: int) -> Tuple[np.array, List[str]]:
    min_agents = defaultdict(lambda: max_workers)
    for node in wg.nodes:
        for req in node.work_unit.worker_reqs:
            min_agents[req.kind] = min(min_agents[req.kind], req.min_count)
    min_counts_list, agent_names = list(zip(*[(count, name) for name, count in min_agents.items()]))
    min_counts: np.array = np.array(min_counts_list, dtype=int)
    return min_counts, agent_names


def get_minimal_counts_by_schedule(scheduled_works: Schedule, agent_names: List[str]) -> np.array:
    workers_intervals = get_workers_intervals(scheduled_works)
    max_used_counts = defaultdict(int)
    for name_index in workers_intervals:
        name, index = name_index.split(SPLITTER)
        max_used_counts[name] = max(max_used_counts[name], int(index))
    optimal_counts = np.array([max_used_counts[name] + 1 for name in agent_names], dtype=int)
    return optimal_counts


def init_borders(wg: WorkGraph, scheduler: Scheduler, deadline: Time,
                 worker_factor: int,
                 max_workers: int,
                 right_agents: WorkerContractorPool or None) -> Tuple[Optional[np.ndarray], Optional[np.ndarray], List[str]]:
    left_counts, agent_names = find_min_agents(wg, max_workers)
    if right_agents is None:
        right_counts: np.ndarray = left_counts * worker_factor
    else:
        right_counts = np.array([right_agents[name] for name in agent_names])

    if is_resources_good(wg, left_counts, agent_names, scheduler, deadline):
        # wg is resolved in time by minimal set of workers
        return left_counts, None, agent_names

    if not is_resources_good(wg, right_counts, agent_names, scheduler, deadline):
        # wg is not resolved in time by any set of workers
        return None, right_counts, agent_names
    return left_counts, right_counts, agent_names


def prepare_answer(counts: np.array, agent_names: List[str], wg: WorkGraph, scheduler: Scheduler,
                   dry_resources: bool):
    schedule, max_time = get_schedule_with_time(scheduler, wg, counts, agent_names)
    if dry_resources:
        optimal_counts = get_minimal_counts_by_schedule(schedule, agent_names)
    else:
        optimal_counts = counts
    contractor = agents_to_contractors(optimal_counts, agent_names)[0]
    return contractor, max_time



from collections import defaultdict
from dataclasses import dataclass, field
from random import Random
from typing import List, Optional, Callable

from sampo.schemas.identifiable import Identifiable
from sampo.schemas.requirements import WorkerReq, EquipmentReq, MaterialReq, ConstructionObjectReq
from sampo.schemas.resources import Worker
from sampo.schemas.serializable import AutoJSONSerializable
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.serializers import custom_serializer


# TODO: describe the class (description, parameters)
@dataclass
class WorkUnit(AutoJSONSerializable['WorkUnit'], Identifiable):
    worker_reqs: List[WorkerReq] = field(default_factory=list)
    equipment_reqs: List[EquipmentReq] = field(default_factory=list)
    material_reqs: List[MaterialReq] = field(default_factory=list)
    object_reqs: List[ConstructionObjectReq] = field(default_factory=list)
    group: Optional[str] = "default"
    is_service_unit: Optional[bool] = False
    # TODO Remove optional
    volume: Optional[float] = 1
    volume_type: Optional[str] = "unit"

    # TODO: describe the function (description, parameters, return type)
    @custom_serializer('worker_reqs')
    def worker_reqs_serializer(self, value):
        return [wr._serialize() for wr in value]

    # TODO: describe the function (description, parameters, return type)
    @classmethod
    @custom_serializer('worker_reqs', deserializer=True)
    def worker_reqs_deserializer(cls, value):
        return [WorkerReq._deserialize(wr) for wr in value]

    # TODO: describe the function (description, parameters, return type)
    # TODO: move this logit to WorkTimeEstimator
    def estimate_static(self, worker_list: List[Worker], work_estimator: WorkTimeEstimator = None) -> Time:
        if work_estimator:
            # TODO What?? We are loosing much performance
            #  when processing worker names EACH estimation
            workers = {w.name.replace("_res_fact", ""): w.count for w in worker_list}
            work_time = work_estimator.estimate_time(self.name.split("_stage_")[0], self.volume, workers)
            if work_time > 0:
                return work_time

        return self._abstract_estimate(worker_list, get_static_by_worker)

    # TODO: describe the function (description, parameters, return type)
    def estimate_stochastic(self, worker_list: List[Worker], rand: Optional[Random] = None) -> Time:
        return self._abstract_estimate(worker_list, get_stochastic_by_worker, rand)

    # TODO: describe the function (description, parameters, return type)
    def _abstract_estimate(self, worker_list: List[Worker],
                           get_productivity: Callable[[Worker, Optional[Random]], float],
                           rand: Optional[Random] = None) -> Time:
        groups = defaultdict(list)
        for w in worker_list:
            groups[w.name].append(w)
        times = [Time(0)]  # if there are no requirements for the work, it is done instantly
        for req in self.worker_reqs:
            if req.min_count == 0:
                continue
            name = req.kind
            command = groups[name]
            worker_count = sum([worker.count for worker in command], 0)
            if worker_count < req.min_count:
                return Time.inf()
            productivity = sum([get_productivity(c, rand) for c in command], 0) / worker_count
            productivity *= communication_coefficient(worker_count, req.max_count)
            if productivity == 0:
                return Time.inf()
            times.append(Time(req.volume // productivity))
        return max(times)

    def __getstate__(self):
        # custom method to avoid calling __hash__() on GraphNode objects
        return self._serialize()

    def __setstate__(self, state):
        new_work_unit = self._deserialize(state)
        self.worker_reqs = new_work_unit.worker_reqs
        self.equipment_reqs = new_work_unit.equipment_reqs
        self.object_reqs = new_work_unit.object_reqs
        self.material_reqs = new_work_unit.material_reqs
        self.id = new_work_unit.id
        self.name = new_work_unit.name
        self.is_service_unit = new_work_unit.is_service_unit
        self.volume = new_work_unit.volume
        self.volume_type = new_work_unit.volume_type
        self.group = new_work_unit.group


# TODO: describe the function (description, parameters, return type)
def get_static_by_worker(w: Worker, _: Optional[Random] = None):
    return w.get_static_productivity()


# TODO: describe the function (description, parameters, return type)
def get_stochastic_by_worker(w: Worker, rand: Optional[Random] = None):
    return w.get_stochastic_productivity(rand)


# Function is chosen because it has a quadratic decrease in efficiency as the number of commands on the object
# increases, after the maximum number of commands begins to decrease in efficiency, and its growth rate depends on
# the maximum number of commands.
# sum(1 - ((x-1)^2 / max_groups^2), where x from 1 to groups_count
# TODO: describe the function (description, parameters, return type)
def communication_coefficient(groups_count: int, max_groups: int) -> float:
    n = groups_count
    m = max_groups
    return 1 / (6 * m ** 2) * (-2 * n ** 3 + 3 * n ** 2 + (6 * m ** 2 - 1) * n)



import random
from abc import ABC, abstractmethod
from functools import partial
from typing import List, Dict, Iterable, Callable

import numpy as np
from deap import creator, base, tools
from deap.base import Toolbox

from sampo.scheduler.genetic.converter import convert_chromosome_to_schedule
from sampo.scheduler.genetic.converter import convert_schedule_to_chromosome, ChromosomeType
from sampo.scheduler.topological.base import RandomizedTopologicalScheduler
from sampo.schemas.contractor import Contractor, WorkerContractorPool, get_worker_contractor_pool
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import ScheduledWork
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.collections_util import reverse_dictionary

# logger = mp.log_to_stderr(logging.DEBUG)


class FitnessFunction(ABC):

    def __init__(self, evaluator: Callable[[list[ChromosomeType]], list[int]]):
        self._evaluator = evaluator

    @abstractmethod
    def evaluate(self, chromosomes: list[ChromosomeType]) -> list[int]:
        ...


class TimeFitness(FitnessFunction):

    def __init__(self, evaluator: Callable[[list[ChromosomeType]], list[int]]):
        super().__init__(evaluator)

    def evaluate(self, chromosomes: list[ChromosomeType]) -> list[int]:
        return self._evaluator(chromosomes)


class TimeAndResourcesFitness(FitnessFunction):

    def __init__(self, evaluator: Callable[[list[ChromosomeType]], list[int]]):
        super().__init__(evaluator)

    def evaluate(self, chromosomes: list[ChromosomeType]) -> list[int]:
        evaluated = self._evaluator(chromosomes)
        return [finish_time + int(np.sum(chromosome[2])) for finish_time, chromosome in zip(evaluated, chromosomes)]


class DeadlineResourcesFitness(FitnessFunction):

    def __init__(self, deadline: Time, evaluator: Callable[[list[ChromosomeType]], list[int]]):
        super().__init__(evaluator)
        self._deadline = deadline

    @staticmethod
    def prepare(deadline: Time):
        """
        Returns the constructor of that fitness function prepared to use in Genetic
        """
        return partial(DeadlineResourcesFitness, deadline)

    def evaluate(self, chromosomes: list[ChromosomeType]) -> list[int]:
        evaluated = self._evaluator(chromosomes)
        return [Time.inf() if finish_time > self._deadline else int(np.sum(chromosome[2])) for finish_time, chromosome in zip(evaluated, chromosomes)]


# create class FitnessMin, the weights = -1 means that fitness - is function for minimum

creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
creator.create("Individual", list, fitness=creator.FitnessMin)
Individual = creator.Individual


def init_toolbox(wg: WorkGraph, contractors: List[Contractor], worker_pool: WorkerContractorPool,
                 index2node: Dict[int, GraphNode],
                 work_id2index: Dict[str, int], worker_name2index: Dict[str, int],
                 index2contractor: Dict[int, str],
                 index2contractor_obj: Dict[int, Contractor],
                 init_chromosomes: Dict[str, ChromosomeType],
                 mutate_order: float, mutate_resources: float, selection_size: int,
                 rand: random.Random,
                 spec: ScheduleSpec,
                 worker_pool_indices: dict[int, dict[int, Worker]],
                 contractor2index: dict[str, int],
                 contractor_borders: np.ndarray,
                 node_indices: list[int],
                 index2node_list: list[tuple[int, GraphNode]],
                 parents: Dict[int, list[int]],
                 assigned_parent_time: Time = Time(0),
                 work_estimator: WorkTimeEstimator = None) -> base.Toolbox:
    toolbox = base.Toolbox()
    # generate initial population
    toolbox.register("generate_chromosome", generate_chromosome, wg=wg, contractors=contractors,
                     index2node_list=index2node_list, work_id2index=work_id2index, worker_name2index=worker_name2index,
                     contractor2index=contractor2index, contractor_borders=contractor_borders,
                     init_chromosomes=init_chromosomes, rand=rand, work_estimator=work_estimator)

    # create from generate_chromosome function one individual
    toolbox.register("individual", tools.initRepeat, Individual, toolbox.generate_chromosome, n=1)
    # create population from individuals
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    # crossover for order
    toolbox.register("mate", mate_scheduling_order, rand=rand)
    # mutation for order. Coefficient luke one or two mutation in individual
    toolbox.register("mutate", tools.mutShuffleIndexes, indpb=mutate_order)
    # selection. Some random individuals and arranges a battle between them as a result in a continuing genus,
    # this is the best among these it
    toolbox.register("select", tools.selTournament, tournsize=selection_size)

    # mutation for resources
    toolbox.register("mutate_resources", mut_uniform_int, probability_mutate_resources=mutate_resources,
                     contractor_count=len(index2contractor), rand=rand)
    # mutation for resource borders
    toolbox.register("mutate_resource_borders", mutate_resource_borders,
                     probability_mutate_contractors=mutate_resources, rand=rand)
    # crossover for resources
    toolbox.register("mate_resources", mate_for_resources, rand=rand)
    # crossover for resource borders
    toolbox.register("mate_resource_borders", mate_for_resource_borders, rand=rand)

    toolbox.register("validate", is_chromosome_correct, node_indices=node_indices, parents=parents)
    toolbox.register("schedule_to_chromosome", convert_schedule_to_chromosome, wg=wg,
                     work_id2index=work_id2index, worker_name2index=worker_name2index,
                     contractor2index=contractor2index, contractor_borders=contractor_borders)
    toolbox.register("chromosome_to_schedule", convert_chromosome_to_schedule, worker_pool=worker_pool,
                     index2node=index2node, index2contractor=index2contractor_obj,
                     worker_pool_indices=worker_pool_indices, spec=spec, assigned_parent_time=assigned_parent_time,
                     work_estimator=work_estimator)
    return toolbox


def copy_chromosome(c: ChromosomeType) -> ChromosomeType:
    return c[0].copy(), c[1].copy(), c[2].copy()


def generate_chromosome(wg: WorkGraph, contractors: List[Contractor], index2node_list: list[tuple[int, GraphNode]],
                        work_id2index: Dict[str, int], worker_name2index: Dict[str, int],
                        contractor2index: Dict[str, int], contractor_borders: np.ndarray,
                        init_chromosomes: Dict[str, ChromosomeType], rand: random.Random,
                        work_estimator: WorkTimeEstimator = None) -> ChromosomeType:
    """
    It is necessary to generate valid scheduling, which are satisfied to current dependencies
    That's why will be used the approved order of works (HEFT order and Topological sorting)
    Topological sorts are generating always different
    HEFT is always the same(now not)
    HEFT we will choose in 30% of attempts
    Topological in others

    :param work_estimator:
    :param contractors:
    :param wg:
    :param work_id2index:
    :param index2node_list:
    :param worker_name2index:
    :param contractor2index:
    :param contractor_borders:
    :param rand:
    :param init_chromosomes:
    :return: chromosome
    """
    chance = rand.random()
    if chance < 0.2:
        chromosome = init_chromosomes["heft_end"]
    elif chance < 0.4:
        chromosome = init_chromosomes["heft_between"]
    else:
        schedule = RandomizedTopologicalScheduler(work_estimator,
                                                  int(rand.random() * 1000000)) \
            .schedule(wg, contractors)
        chromosome = convert_schedule_to_chromosome(wg, work_id2index, worker_name2index,
                                                    contractor2index, contractor_borders, schedule)
    return chromosome


def is_chromosome_correct(chromosome: ChromosomeType,
                          node_indices: list[int],
                          parents: Dict[int, list[int]]) -> bool:
    return is_chromosome_order_correct(chromosome, parents) and \
           is_chromosome_contractors_correct(chromosome, node_indices)


def is_chromosome_order_correct(chromosome: ChromosomeType, parents: Dict[int, list[int]]) -> bool:
    work_order = chromosome[0]
    used = set()
    for work_index in work_order:
        used.add(work_index)
        for parent in parents[work_index]:
            if parent not in used:
                # logger.error(f'Order validation failed: {work_order}')
                return False
    return True


def is_chromosome_contractors_correct(chromosome: ChromosomeType,
                                      work_indices: Iterable[int]) -> bool:
    """
    Checks that assigned contractors can supply assigned workers

    :param chromosome:
    :param contractors_borders:
    :param work_indices:
    :return:
    """
    for work_ind in work_indices:
        resources_count = chromosome[1][work_ind, :-1]
        contractor_ind = chromosome[1][work_ind, -1]
        contractor_border = chromosome[2][contractor_ind]
        for ind, count in enumerate(resources_count):
            if contractor_border[ind] < count:
                # logger.error(f'Contractor border validation failed: {contractor_border[ind]} < {count}')
                return False
    return True


def get_order_tail(head_set: np.ndarray, other: np.ndarray) -> np.ndarray:
    head_set = set(head_set)
    return np.array([node for node in other if node not in head_set])


def mate_scheduling_order(ind1: ChromosomeType, ind2: ChromosomeType, rand: random.Random) \
        -> (ChromosomeType, ChromosomeType):
    """
    Crossover for order
    Basis crossover is cxOnePoint
    But we checked not repeated works in individual order

    :param ind1:
    :param ind2:
    :param rand:
    :return: two cross individuals
    """
    ind1 = copy_chromosome(ind1)
    ind2 = copy_chromosome(ind2)

    order1 = ind1[0]
    order2 = ind2[0]

    # randomly select the point where the crossover will take place
    crossover_point = rand.randint(1, len(ind1))

    ind1_new_tail = get_order_tail(order1[:crossover_point], order2)
    ind2_new_tail = get_order_tail(order2[:crossover_point], order1)

    order1[crossover_point:] = ind1_new_tail
    order2[crossover_point:] = ind2_new_tail

    return ind1, ind2


def mut_uniform_int(ind: ChromosomeType, low: np.ndarray, up: np.ndarray, type_of_worker: int,
                    probability_mutate_resources: float, contractor_count: int, rand: random.Random) -> ChromosomeType:
    """
    Mutation function for resources
    It changes selected numbers of workers in random work in certain interval for this work

    :param contractor_count:
    :param ind:
    :param low: lower bound specified by `WorkUnit`
    :param up: upper bound specified by `WorkUnit`
    :param type_of_worker:
    :param probability_mutate_resources:
    :param rand:
    :return: mutate individual
    """
    ind = copy_chromosome(ind)

    # select random number from interval from min to max from uniform distribution
    size = len(ind[1])
    workers_count = len(ind[1][0])

    if type_of_worker == workers_count - 1:
        # print('Contractor mutation!')
        for i in range(size):
            if rand.random() < probability_mutate_resources:
                ind[1][i][type_of_worker] = rand.randint(0, contractor_count - 1)
        return ind

    # change in this interval in random number from interval
    for i, xl, xu in zip(range(size), low, up):
        if rand.random() < probability_mutate_resources:
            # borders
            contractor = ind[1][i][-1]
            border = ind[2][contractor][type_of_worker]
            # TODO Debug why min(xu, border) can be lower than xl
            ind[1][i][type_of_worker] = rand.randint(xl, max(xl, min(xu, border)))

    return ind


def mutate_resource_borders(ind: ChromosomeType, contractors_capacity: np.ndarray, resources_min_border: np.ndarray,
                            type_of_worker: int, probability_mutate_contractors: float, rand: random.Random) \
        -> ChromosomeType:
    """
    Mutation for contractors' resource borders.

    :param ind:
    :param contractors_capacity:
    :param resources_min_border:
    :param type_of_worker:
    :param probability_mutate_contractors:
    :param rand:
    :return:
    """
    ind = copy_chromosome(ind)

    num_contractors = len(ind[2])
    for i in range(num_contractors):
        if rand.random() < probability_mutate_contractors:
            ind[2][i][type_of_worker] -= rand.randint(resources_min_border[type_of_worker] + 1,
                                                      max(resources_min_border[type_of_worker] + 1,
                                                          ind[2][i][type_of_worker] // 10))
            if ind[2][i][type_of_worker] <= 0:
                ind[2][i][type_of_worker] = 1

    return ind


def mate_for_resources(ind1: ChromosomeType, ind2: ChromosomeType, mate_positions: np.ndarray,
                       rand: random.Random) -> (ChromosomeType, ChromosomeType):
    """
    CxOnePoint for resources

    :param ind1: first individual
    :param ind2: second individual
    :param mate_positions: an array of positions that should be mate
    :param rand: the rand object used for exchange point selection
    :return: first and second individual
    """
    ind1 = copy_chromosome(ind1)
    ind2 = copy_chromosome(ind2)

    # exchange work resources
    res1 = ind1[1][:, mate_positions]
    res2 = ind2[1][:, mate_positions]
    cxpoint = rand.randint(1, len(res1))

    mate_positions = rand.sample(list(range(len(res1))), cxpoint)

    res1[mate_positions], res2[mate_positions] = res2[mate_positions], res1[mate_positions]
    return ind1, ind2


def mate_for_resource_borders(ind1: ChromosomeType, ind2: ChromosomeType,
                              mate_positions: np.ndarray, rand: random.Random) -> (ChromosomeType, ChromosomeType):
    ind1 = copy_chromosome(ind1)
    ind2 = copy_chromosome(ind2)

    num_contractors = len(ind1[2])
    contractors_to_mate = rand.sample(list(range(num_contractors)), rand.randint(1, num_contractors))

    if rand.randint(0, 2) == 0:
        # trying to mate whole contractors
        border1 = ind1[2][contractors_to_mate]
        border2 = ind2[2][contractors_to_mate]
        border1[:], border2[:] = border2[:], border1[:]
    else:
        # trying to mate part of selected contractors
        border1 = ind1[2][contractors_to_mate]
        border2 = ind2[2][contractors_to_mate]
        for c_border1, c_border2 in zip(border1, border2):
            # mate_positions = rand.sample(list(range(len(c_border1))), rand.randint(1, len(c_border1)))
            c_border1[mate_positions], c_border2[mate_positions] = c_border2[mate_positions], c_border1[mate_positions]

    return ind1, ind2



from abc import ABC, abstractmethod
from typing import Optional, Callable

from sampo.schemas.contractor import WorkerContractorPool
from sampo.schemas.time import Time


class WorkTimeEstimator(ABC):
    @abstractmethod
    def set_mode(self, use_idle: Optional[bool] = True, mode: Optional[str] = 'realistic'):
        ...

    @abstractmethod
    def estimate_time(self, work_name: str, work_volume: float, resources: WorkerContractorPool) -> Time:
        ...

    @abstractmethod
    def split_to_constructor_and_params(self) -> tuple[Callable[[...], 'WorkTimeEstimator'], tuple]:
        ...


# TODO add simple work_time_estimator based on WorkUnit.estimate_static



from collections import deque
from typing import Dict, List, Tuple, Optional, Union, Iterable

from sortedcontainers import SortedList

from sampo.scheduler.timeline.base import Timeline
from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.graph import GraphNode
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.resources import Worker
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.schemas.types import AgentId, ScheduleEvent, EventType
from sampo.utilities.collections_util import build_index


class MomentumTimeline(Timeline):

    def __init__(self, tasks: Iterable[GraphNode], contractors: Iterable[Contractor], worker_pool: WorkerContractorPool):
        """
        This should create empty Timeline from given list of tasks and contractor list

        :param tasks:
        :param contractors:
        :return:
        """

        # using  time, seq_id and event_type we can guarantee that
        # there may be only one possible order in cases:
        # (a) when events have the same time
        # (in this cases we need both time and seq_id to properly handle
        # available_workers processing logic)
        # (b) when events have the same time and their start and end matches
        # (service tasks for instance may have zero length)
        def event_cmp(x: Union[ScheduleEvent, Time, Tuple[Time, int, int]]) -> Tuple[Time, int, int]:
            if isinstance(x, ScheduleEvent):
                if x.event_type is EventType.Initial:
                    return Time(-1), -1, x.event_type.priority

                return x.time, x.seq_id, x.event_type.priority

            if isinstance(x, Time):
                return x, len(tasks), 1

            if isinstance(x, tuple):
                return x

            raise ValueError(f"Incorrect type of value: {type(x)}")

        # to efficiently search for time slots for tasks to be scheduled
        # we need to keep track of starts and ends of previously scheduled tasks
        # and remember how many workers of a certain type is available at this particular moment
        self._timeline: Dict[str, Dict[str, SortedList[ScheduleEvent]]] = {
            c.id: {
                w_name: SortedList(
                    iterable=(ScheduleEvent(-1, EventType.Initial, Time(0), None, ws.count),),
                    key=event_cmp
                )
                for w_name, ws in c.workers.items()
            }
            for c in contractors
        }

    def find_min_start_time_with_additional(self,
                                            node: GraphNode,
                                            worker_team: List[Worker],
                                            node2swork: Dict[GraphNode, ScheduledWork],
                                            assigned_start_time: Optional[Time] = None,
                                            assigned_parent_time: Time = Time(0),
                                            work_estimator: Optional[WorkTimeEstimator] = None) \
            -> Tuple[Time, Time, Dict[GraphNode, Tuple[Time, Time]]]:
        """
        Computes start time, max parent time, contractor and exec times for given node

        :param worker_team: list of passed workers. Should be IN THE SAME ORDER AS THE CORRESPONDING WREQS
        :param node:
        :param node2swork:
        :param assigned_start_time:
        :param assigned_parent_time:
        :param work_estimator:
        :return: start time, end time, exec_times
        """
        inseparable_chain = node.get_inseparable_chain_with_self()
        contractor_id = worker_team[0].contractor_id if worker_team else ""
        # 1. identify earliest possible start time by max parent's end time

        def apply_time_spec(time: Time):
            return max(time, assigned_start_time) if assigned_start_time is not None else time

        max_parent_time: Time = max(apply_time_spec(
            max((node2swork[pnode].finish_time for pnode in node.parents), default=Time(0))
        ), assigned_parent_time)
        nodes_max_parent_times: Dict[GraphNode, Time] = {n: max((max(apply_time_spec(node2swork[pnode].finish_time),
                                                                     assigned_parent_time)
                                                                 if pnode in node2swork else assigned_parent_time
                                                                 for pnode in n.parents),
                                                                default=assigned_parent_time)
                                                         for n in inseparable_chain}

        # 2. calculating execution time of the task

        exec_time: Time = Time(0)
        exec_times: Dict[GraphNode, Tuple[Time, Time]] = {}  # node: (lag, exec_time)
        for i, chain_node in enumerate(inseparable_chain):
            passed_agents_new = [agent.copy() for agent in worker_team]

            node_exec_time: Time = Time(0) if len(chain_node.work_unit.worker_reqs) == 0 else \
                chain_node.work_unit.estimate_static(passed_agents_new, work_estimator)
            lag_req = nodes_max_parent_times[chain_node] - max_parent_time  # - exec_time
            lag = lag_req if lag_req > 0 else 0

            exec_times[chain_node] = lag, node_exec_time
            exec_time += lag + node_exec_time

        if len(worker_team) == 0:
            return max_parent_time, max_parent_time, exec_times

        st = assigned_start_time if assigned_start_time is not None else self._find_min_start_time(
            self._timeline[contractor_id], inseparable_chain, max_parent_time, exec_time, worker_team
        )

        assert st >= assigned_parent_time

        return st, st + exec_time, exec_times

    def _find_min_start_time(self,
                             resource_timeline: Dict[str, SortedList[ScheduleEvent]],
                             inseparable_chain: List[GraphNode],
                             parent_time: Time,
                             exec_time: Time,
                             passed_agents: List[Worker]) -> Time:
        # if it is a service unit, than it can be satisfied by any contractor at any moment
        # because no real workers is going to be used to execute the task
        # however, we still should respect dependencies of the service task
        # and should start it only after all the dependencies tasks are done
        if all((node.work_unit.is_service_unit for node in inseparable_chain)):
            return parent_time

        # checking if the contractor can satisfy requirements for the task at all
        # we return None in cases when the task cannot be executed
        # even if it is scheduled to the very end, e.g. after the end of all other tasks
        # already scheduled to this contractor

        for node in inseparable_chain:
            for i, wreq in enumerate(node.work_unit.worker_reqs):
                initial_event: ScheduleEvent = resource_timeline[wreq.kind][0]
                assert initial_event.event_type is EventType.Initial
                # if this contractor initially has fewer workers of this type, then needed...
                if initial_event.available_workers_count < passed_agents[i].count:
                    return Time.inf()

        # here we look for the earliest time slot that can satisfy all the worker's specializations
        # we do it in that manner because each worker specialization can be treated separately
        # e.g. requested for different tasks
        # We check only the first node since all inseparable nodes have same worker_reqs despite the difference in exec time
        queue = deque(inseparable_chain[0].work_unit.worker_reqs)

        start = parent_time
        scheduled_wreqs: List[WorkerReq] = []

        type2count: Dict[str, int] = build_index(passed_agents, lambda w: w.name, lambda w: w.count)

        i = 0
        while len(queue) > 0:
            # if i > 0 and i % 50 == 0:
            #     print(f"Warning! Probably cycle in looking for diff workers: {i} iteration")
            i += 1

            wreq = queue.popleft()
            state = resource_timeline[wreq.kind]
            # we look for the earliest time slot starting from 'start' time moment
            # if we have found a time slot for the previous task,
            # we should start to find for the earliest time slot of other task since this new time
            found_start = self._find_earliest_time_slot(state, start, exec_time, type2count[wreq.kind])

            assert found_start >= start

            if len(scheduled_wreqs) == 0 or start == found_start:
                # we schedule the first worker's specialization or the next spec has the same start time
                # as the all previous ones
                scheduled_wreqs.append(wreq)
                start = max(found_start, start)
            else:
                # The current worker specialization can be started only later than
                # the previously found start time.
                # In this case we need to add back all previously scheduled wreq-s into the queue
                # to be scheduled again with the new start time (e.g. found start).
                # This process should reach its termination at least at the very end of this contractor's schedule.
                queue.extend(scheduled_wreqs)
                scheduled_wreqs.clear()
                scheduled_wreqs.append(wreq)
                start = max(found_start, start)

        return start

    @staticmethod
    def _find_earliest_time_slot(state: SortedList[ScheduleEvent],
                                 parent_time: Time,
                                 exec_time: Time,
                                 required_worker_count: int) -> Time:
        current_start_time = parent_time
        current_start_idx = state.bisect_right(current_start_time) - 1

        # the condition means we have reached the end of schedule for this contractor subject to specialization (wreq)
        # as long as we assured that this contractor has enough capacity at all to handle the the task
        # we can stop and put the task at the very end
        i = 0
        while len(state[current_start_idx:]) > 0:
            # if i > 0 and i % 50 == 0:
            #     print(f"Warning! Probably cycle in looking for earliest time slot: {i} iteration")
            #     print(f"Current start time: {current_start_time}, current start idx: {current_start_idx}")
            i += 1
            end_idx = state.bisect_right(current_start_time + exec_time)

            # checking from the end of execution interval, i.e., end_idx - 1
            # up to (including) the event right prepending the start of the execution interval, i.e., current_start_idx - 1
            # we need to check the event current_start_idx - 1 cause it is the first event
            # that influence amount of available for us workers
            not_enough_workers_found = False
            for idx in range(end_idx - 1, current_start_idx - 2, -1):
                if state[idx].available_workers_count < required_worker_count or state[idx].time < parent_time:
                    # we're trying to find a new slot that would start with
                    # either the last index passing the quantity check
                    # or the index after the execution interval
                    # we need max here to process a corner case when the problem arises
                    # on current_start_idx - 1
                    # without max it would get into infinite cycle
                    current_start_idx = max(idx, current_start_idx) + 1
                    not_enough_workers_found = True
                    break

            if not not_enough_workers_found:
                break

            if current_start_idx >= len(state):
                break

            current_start_time = state[current_start_idx].time

        return current_start_time

    def update_timeline(self,
                        task_index: int,
                        finish_time: Time,
                        node: GraphNode,
                        node2swork: Dict[GraphNode, ScheduledWork],
                        worker_team: List[Worker]):
        """
        Inserts `chosen_workers` into the timeline with it's `inseparable_chain`
        :param task_index:
        :param finish_time:
        :param node:
        :param node2swork:
        :param worker_team:
        :return:
        """
        # 7. for each worker's specialization of the chosen contractor being used by the task
        # we update counts of available workers on previously scheduled events
        # that lie between start and end of the task
        # Also, add events of the start and the end to worker's specializations
        # of the chosen contractor.

        # experimental logics lightening. debugging showed its efficiency.

        swork = node2swork[node]  # masking the whole chain ScheduleEvent with the first node
        start = swork.start_time
        end = node2swork[node.get_inseparable_chain_with_self()[-1]].finish_time
        for w in worker_team:
            state = self._timeline[w.contractor_id][w.name]
            start_idx = state.bisect_right(start)
            end_idx = state.bisect_right(end)
            available_workers_count = state[start_idx - 1].available_workers_count
            # updating all events in between the start and the end of our current task
            for event in state[start_idx: end_idx]:
                assert event.available_workers_count >= w.count
                event.available_workers_count -= w.count

            # assert available_workers_count >= w.count

            if start_idx < end_idx:
                event: ScheduleEvent = state[end_idx - 1]
                assert state[0].available_workers_count >= event.available_workers_count + w.count
                end_count = event.available_workers_count + w.count
            else:
                assert state[0].available_workers_count >= available_workers_count
                end_count = available_workers_count

            state.add(ScheduleEvent(task_index, EventType.Start, start, swork, available_workers_count - w.count))
            state.add(ScheduleEvent(task_index, EventType.End, end, swork, end_count))

    def schedule(self,
                 task_index: int,
                 node: GraphNode,
                 node2swork: Dict[GraphNode, ScheduledWork],
                 workers: List[Worker],
                 contractor: Contractor,
                 assigned_start_time: Optional[Time] = None,
                 assigned_time: Optional[Time] = None,
                 assigned_parent_time: Time = Time(0),
                 work_estimator: Optional[WorkTimeEstimator] = None):
        inseparable_chain = node.get_inseparable_chain_with_self()
        st, _, exec_times = \
            self.find_min_start_time_with_additional(node, workers, node2swork, assigned_start_time,
                                                     assigned_parent_time, work_estimator)
        if assigned_time is not None:
            exec_times = {n: (Time(0), assigned_time // len(inseparable_chain))
                          for n in inseparable_chain}

        # TODO Decide how to deal with exec_times(maybe we should remove using pre-computed exec_times)
        self._schedule_with_inseparables(task_index, node, node2swork, inseparable_chain,
                                         workers, contractor, st, exec_times)

    def _schedule_with_inseparables(self,
                                    task_index: int,
                                    node: GraphNode,
                                    node2swork: Dict[GraphNode, ScheduledWork],
                                    inseparable_chain: List[GraphNode],
                                    worker_team: List[Worker],
                                    contractor: Contractor,
                                    start_time: Time,
                                    exec_times: Dict[GraphNode, Tuple[Time, Time]]):
        # 6. create a schedule entry for the task

        nodes_start_times: Dict[GraphNode, Time] = {n: max((node2swork[pnode].finish_time
                                                            if pnode in node2swork else Time(0)
                                                            for pnode in n.parents),
                                                           default=Time(0))
                                                    for n in inseparable_chain}

        curr_time = start_time
        for i, chain_node in enumerate(inseparable_chain):
            _, node_time = exec_times[chain_node]

            # TODO What?)
            # lag_req = nodes_start_times[chain_node] - start_time - node_time
            lag_req = nodes_start_times[chain_node] - curr_time
            node_lag = lag_req if lag_req > 0 else 0

            start_work = curr_time + node_lag
            swork = ScheduledWork(
                work_unit=chain_node.work_unit,
                start_end_time=(start_work, start_work + node_time),
                workers=worker_team,
                contractor=contractor
            )
            curr_time += node_time + node_lag
            node2swork[chain_node] = swork

        self.update_timeline(task_index, curr_time, node, node2swork, worker_team)

    def __getitem__(self, item: AgentId):
        return self._timeline[item[0]][item[1]]



from sampo.pipeline.base import InputPipeline, SchedulePipeline
from sampo.pipeline.delegating import DelegatingScheduler
from sampo.scheduler.base import Scheduler
from sampo.scheduler.generic import GenericScheduler
from sampo.scheduler.utils.local_optimization import OrderLocalOptimizer, ScheduleLocalOptimizer
from sampo.schemas.apply_queue import ApplyQueue
from sampo.schemas.contractor import Contractor, get_worker_contractor_pool
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator


class DefaultInputPipeline(InputPipeline):

    def __init__(self):
        self._wg = None
        self._contractors = None
        self._work_estimator = None
        self._node_order = None
        self._spec = ScheduleSpec()
        self._assigned_parent_time = Time(0)
        self._local_optimize_stack = ApplyQueue()

    def wg(self, wg: WorkGraph) -> 'InputPipeline':
        self._wg = wg
        return self

    def contractors(self, contractors: list[Contractor]) -> 'InputPipeline':
        self._contractors = contractors
        return self

    def spec(self, spec: ScheduleSpec) -> 'InputPipeline':
        self._spec = spec
        return self

    def time_shift(self, time: Time) -> 'InputPipeline':
        self._assigned_parent_time = time
        return self

    def work_estimator(self, work_estimator: WorkTimeEstimator) -> 'InputPipeline':
        self._work_estimator = work_estimator
        return self

    def node_order(self, node_order: list[GraphNode]) -> 'InputPipeline':
        self._node_order = node_order
        return self

    def optimize_local(self, optimizer: OrderLocalOptimizer, area: range) -> 'InputPipeline':
        self._local_optimize_stack.add(optimizer.optimize, (area,))
        return self

    def schedule(self, scheduler: Scheduler) -> 'SchedulePipeline':
        if isinstance(scheduler, GenericScheduler):
            # if scheduler is generic, it supports injecting local optimizations
            s_self = self  # cache upper-layer self to another variable to get it from inner class

            class LocalOptimizedScheduler(DelegatingScheduler):

                def __init__(self, delegate: GenericScheduler):
                    super().__init__(delegate)

                def delegate_prioritization(self, orig_prioritization):
                    def prioritization(wg: WorkGraph, work_estimator: WorkTimeEstimator):
                        # call delegate's prioritization and apply local optimizations
                        return s_self._local_optimize_stack.apply(orig_prioritization(wg, work_estimator))

                    return prioritization

            scheduler = LocalOptimizedScheduler(scheduler)
        elif not self._local_optimize_stack.empty():
            print('Trying to apply local optimizations to non-generic scheduler, ignoring it')

        schedule, _, _, node_order = scheduler.schedule_with_cache(self._wg, self._contractors, self._spec,
                                                                   assigned_parent_time=self._assigned_parent_time)
        self._node_order = node_order
        return DefaultSchedulePipeline(self, schedule)


# noinspection PyProtectedMember
class DefaultSchedulePipeline(SchedulePipeline):

    def __init__(self, s_input: DefaultInputPipeline, schedule: Schedule):
        self._input = s_input
        self._worker_pool = get_worker_contractor_pool(s_input._contractors)
        self._schedule = schedule
        self._scheduled_works = {s_input._wg[swork.work_unit.id]:
                                 swork for swork in schedule.to_schedule_work_dict.values()}
        self._local_optimize_stack = ApplyQueue()

    def optimize_local(self, optimizer: ScheduleLocalOptimizer, area: range) -> 'SchedulePipeline':
        self._local_optimize_stack.add(optimizer.optimize,
                                       (self._input._node_order, self._input._contractors, self._input._spec,
                                        self._worker_pool, self._input._work_estimator,
                                        self._input._assigned_parent_time, area))
        return self

    def finish(self) -> Schedule:
        processed_sworks = self._local_optimize_stack.apply(self._scheduled_works)
        return Schedule.from_scheduled_works(processed_sworks.values(), self._input._wg)



from native import decodeEvaluationInfo
from native import evaluate
from native import freeEvaluationInfo

from sampo.scheduler.genetic.converter import ChromosomeType
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.resources import Worker
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.collections_util import reverse_dictionary


class NativeWrapper:
    def __init__(self, wg: WorkGraph, contractors: list[Contractor], worker_name2index: dict[str, int],
                 worker_pool_indices: dict[int, dict[int, Worker]], time_estimator: WorkTimeEstimator):
        # the outer numeration. Begins with inseparable heads, continuous with tails.
        numeration: dict[int, GraphNode] = {i: node for i, node in
                                            enumerate(filter(lambda node: not node.is_inseparable_son(), wg.nodes))}
        heads_count = len(numeration)
        for i, node in enumerate([node for node in wg.nodes if node.is_inseparable_son()]):
            numeration[heads_count + i] = node
        rev_numeration = reverse_dictionary(numeration)

        self.numeration = numeration
        # for each vertex index store list of parents' indices
        self.parents = [[rev_numeration[p] for p in numeration[index].parents] for index in range(wg.vertex_count)]
        # for each vertex index store list of whole it's inseparable chain indices
        self.inseparables = [[rev_numeration[p] for p in numeration[index].get_inseparable_chain_with_self()]
                             for index in range(wg.vertex_count)]
        # contractors' workers matrix. If contractor can't supply given type of worker, 0 should be passed
        self.workers = [[0 for _ in range(len(worker_name2index))] for _ in contractors]
        for i, contractor in enumerate(contractors):
            for worker in contractor.workers.values():
                self.workers[i][worker_name2index[worker.name]] = worker.count

        self.totalWorksCount = wg.vertex_count
        self.time_estimator = time_estimator
        self.worker_pool_indices = worker_pool_indices
        self._current_chromosomes = None

        # preparing C++ cache
        self._cache = decodeEvaluationInfo(self, self.parents, self.inseparables, self.workers, self.totalWorksCount)

    def calculate_working_time(self, chromosome_ind: int, team_target: int, work: int) -> int:
        team = self._current_chromosomes[chromosome_ind][1][team_target]
        workers = [self.worker_pool_indices[worker_index][team[len(self.workers[0])]]
                   .copy().with_count(team[worker_index])
                   for worker_index in range(len(self.workers[0]))
                   if team[worker_index] > 0]
        return self.numeration[work].work_unit.estimate_static(workers, self.time_estimator).value

    def evaluate(self, chromosomes: list[ChromosomeType]):
        self._current_chromosomes = chromosomes
        return evaluate(self._cache, chromosomes)

    def close(self):
        freeEvaluationInfo(self._cache)



from collections import deque
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property, cache
from typing import List, Union, Tuple, Optional, Dict, Set

import numpy as np
from scipy.sparse import dok_matrix

from sampo.schemas.serializable import JSONSerializable, T, JS
from sampo.schemas.works import WorkUnit


# TODO: describe the class (description, parameters)
class EdgeType(Enum):
    InseparableFinishStart = 'IFS'
    LagFinishStart = 'FFS'
    StartStart = 'SS'
    FinishFinish = 'FF'
    FinishStart = 'FS'
    StartFinish = 'SF'


# TODO: describe the class (description, parameters)
@dataclass
class GraphEdge:
    start: 'GraphNode'
    finish: 'GraphNode'
    lag: Optional[float] = 0
    type: Optional[EdgeType] = None


# TODO: describe the class (description, parameters)
class GraphNode(JSONSerializable['GraphNode']):
    _work_unit: WorkUnit
    _parent_edges: List[GraphEdge]
    _children_edges: List[GraphEdge]

    # TODO: describe the function (description, parameters, return type)
    def __init__(self, work_unit: WorkUnit,
                 parent_works: Union[List['GraphNode'], List[Tuple['GraphNode', float, EdgeType]]]):
        self._work_unit = work_unit
        self._parent_edges = []
        self.add_parents(parent_works)
        self._children_edges = []

    def __hash__(self) -> int:
        return hash(self.id)

    def __repr__(self) -> str:
        return self.id

    def __getstate__(self):
        # custom method to avoid calling __hash__() on GraphNode objects
        return self._work_unit._serialize(), \
               [(e.start.id, e.lag, e.type.value) for e in self._parent_edges]

    def __setstate__(self, state):
        # custom method to avoid calling __hash__() on GraphNode objects
        s_work_unit, s_parent_edges = state
        self.__init__(WorkUnit._deserialize(s_work_unit),
                      s_parent_edges)
        # self._work_unit = representation['work_unit']
        # self._parent_edges = [GraphEdge(*e) for e in representation['parent_edges']]

    # TODO: describe the function (description, return type)
    def _serialize(self) -> T:
        return {
            'work_unit': self._work_unit._serialize(),
            'parent_edges': [(e.start.id, e.lag, e.type.value) for e in self._parent_edges],
            # 'child_edges': [(e.finish.work_unit.id, e.lag, e.type.value) for e in self._children_edges]
        }

    # TODO: describe the function (description, parameters, return type)
    @classmethod
    def _deserialize(cls, representation: T) -> dict:
        representation['work_unit'] = WorkUnit._deserialize(representation['work_unit'])
        representation['parent_edges'] = [(e[0], e[1], EdgeType(e[2])) for e in representation['parent_edges']]
        # representation['child_edges'] = [(e[0], e[1], EdgeType(e[2])) for e in representation['child_edges']]
        return representation

    # TODO: describe the function (description, parameters, return type)
    def update_work_unit(self, work_unit: WorkUnit) -> None:
        self._work_unit = work_unit

    # TODO: add a comment about the returned type and complete the description
    def add_parents(self, parent_works: List['GraphNode'] or List[Tuple['GraphNode', float, EdgeType]]) -> None:
        """
        Two-sided linking of children and parents

        :param parent_works:
        """
        edges: List[GraphEdge] = []
        if len(parent_works) > 0 and isinstance(parent_works[0], GraphNode):
            edges = [GraphEdge(p, self, -1, EdgeType.FinishStart) for p in parent_works]
        elif len(parent_works) > 0 and isinstance(parent_works[0], tuple):
            edges = [GraphEdge(p, self, lag, edge_type) for p, lag, edge_type in parent_works]

        for i, p in enumerate(parent_works):
            p: GraphNode = p[0] if isinstance(p, tuple) else p
            p._add_child_edge(edges[i])
        self._parent_edges += edges

    # TODO: describe the function (description, return type)
    def is_inseparable_parent(self) -> bool:
        return self.inseparable_son is not None

    # TODO: describe the function (description, return type)
    def is_inseparable_son(self) -> bool:
        return self.inseparable_parent is not None

    # TODO: describe the function (description, parameters, return type)
    def traverse_children(self, topologically: bool = False):
        visited_vertexes = set()
        vertexes_to_visit = deque([self])
        while len(vertexes_to_visit) > 0:
            v = vertexes_to_visit.popleft()
            if topologically and any(p not in visited_vertexes for p in v.parents):
                vertexes_to_visit.append(v)
                continue
            if v not in visited_vertexes:
                visited_vertexes.add(v)
                vertexes_to_visit.extend(v.children)
                yield v

    # TODO: describe the function (description, return type)
    @cached_property
    def inseparable_son(self) -> Optional['GraphNode']:
        inseparable_children = list([x.finish for x in self._children_edges
                                     if x.type == EdgeType.InseparableFinishStart])
        return inseparable_children[0] if inseparable_children else None

    # TODO: describe the function (description, return type)
    @cached_property
    def inseparable_parent(self) -> Optional['GraphNode']:
        inseparable_parents = list([x.start for x in self._parent_edges if x.type == EdgeType.InseparableFinishStart])
        return inseparable_parents[0] if inseparable_parents else None

    # TODO Describe the function (description, return type)
    @cached_property
    def parents(self) -> List['GraphNode']:
        return list([edge.start for edge in self._parent_edges])

    # TODO Describe the function (description, return type)
    @cached_property
    def parents_set(self) -> Set['GraphNode']:
        return set(self.parents)

    # TODO Describe the function (description, return type)
    @cached_property
    def children(self) -> List['GraphNode']:
        return list([edge.finish for edge in self._children_edges])

    # TODO Describe the function (description, return type)
    @cached_property
    def children_set(self) -> Set['GraphNode']:
        return set(self.children)

    @property
    def edges_to(self) -> List[GraphEdge]:
        return self._parent_edges

    # TODO: describe the function (description, return type)
    @property
    def edges_from(self) -> List[GraphEdge]:
        return self._children_edges

    # TODO: describe the function (description, return type)
    @property
    def work_unit(self) -> WorkUnit:
        return self._work_unit

    # TODO: describe the function (description, return type)
    @property
    def id(self) -> str:
        return self.work_unit.id

    @cache
    def get_inseparable_chain(self) -> Optional[List['GraphNode']]:
        """
        Gets an ordered list of whole chain of nodes, connected with edges of type INSEPARABLE_FINISH_START =
        'INSEPARABLE',
        IF self NODE IS THE START NODE OF SUCH CHAIN. Otherwise, None.

        :return: List of GraphNode or None
        """
        return [self] + self._get_inseparable_children() \
            if bool(self.inseparable_son) and not bool(self.inseparable_parent) \
            else None

    def get_inseparable_chain_with_self(self) -> List['GraphNode']:
        """
        Gets an ordered list of whole chain of nodes, connected with edges of type INSEPARABLE_FINISH_START =
        'INSEPARABLE'.

        :return: List of `inseparable chain` with starting node
        """
        return self.get_inseparable_chain() if self.get_inseparable_chain() else [self]

    def _get_inseparable_children(self) -> List['GraphNode']:
        """
        Recursively gets a child, connected with INSEPARABLE_FINISH_START edge, its inseparable child, etc.
        As any node may contain an inseparable connection with only one of its children, there is no need to choose.
        If no children are connected inseparably, returns None.

        :return: List[GraphNode]. Empty, if there is no inseparable children
        """
        inseparable_child = self.inseparable_son
        return [inseparable_child] + inseparable_child._get_inseparable_children() \
            if inseparable_child \
            else []

    # TODO: describe the function (description, parameters, return type)
    def _add_child_edge(self, child: GraphEdge):
        self._children_edges.append(child)


GraphNodeDict = Dict[str, GraphNode]


# TODO: describe the class (description, parameters)
# TODO Make property for list of GraphEdges??
@dataclass(frozen=True)
class WorkGraph(JSONSerializable['WorkGraph']):
    start: GraphNode
    finish: GraphNode

    nodes: List[GraphNode] = field(init=False)
    adj_matrix: dok_matrix = field(init=False)
    dict_nodes: GraphNodeDict = field(init=False)
    vertex_count: int = field(init=False)

    # TODO: describe the function (description, parameters, return type)
    def __post_init__(self) -> None:
        ordered_nodes, adj_matrix, dict_nodes = self._to_adj_matrix()
        # To avoid field set of frozen instance errors
        object.__setattr__(self, 'nodes', ordered_nodes)
        object.__setattr__(self, 'adj_matrix', adj_matrix)
        object.__setattr__(self, 'dict_nodes', dict_nodes)
        object.__setattr__(self, 'vertex_count', len(ordered_nodes))

    def __hash__(self):
        return hash(self.start) + 17 * hash(self.finish)

    def __getitem__(self, item: str) -> GraphNode:
        return self.dict_nodes[item]

    def __getstate__(self):
        # custom method to avoid calling __hash__() on GraphNode objects
        representation = self._serialize()
        representation['start'] = self.start.id
        representation['finish'] = self.start.id
        return representation

    def __setstate__(self, state):
        # custom method to avoid calling __hash__() on GraphNode objects
        deserialized = self._deserialize(state)
        object.__setattr__(self, 'start', deserialized.start)
        object.__setattr__(self, 'finish', deserialized.finish)
        self.__post_init__()

    # TODO: describe the function (description, return type)
    def _serialize(self) -> T:
        return {
            'nodes': [graph_node._serialize() for graph_node in self.nodes]
        }

    # TODO: describe the function (description, parameters, return type)
    @classmethod
    def _deserialize(cls, representation: T) -> JS:
        serialized_nodes = [GraphNode._deserialize(node) for node in representation['nodes']]
        assert not serialized_nodes[0]['parent_edges']
        start_id, finish_id = (serialized_nodes[i]['work_unit'].id for i in (0, -1))

        nodes_dict = dict()
        for node_info in serialized_nodes:
            wu, parent_info = (node_info[member] for member in ('work_unit', 'parent_edges'))
            graph_node = GraphNode(wu, [(nodes_dict[p_id], p_lag, p_type) for p_id, p_lag, p_type in parent_info])
            nodes_dict[wu.id] = graph_node

        return WorkGraph(nodes_dict[start_id], nodes_dict[finish_id])

    # TODO: describe the function (description, return type)
    # TODO Check that adj matrix is really need
    def _to_adj_matrix(self) -> Tuple[List[GraphNode], dok_matrix, Dict[str, GraphNode]]:
        ordered_nodes: List[GraphNode] = list(self.start.traverse_children(topologically=True))
        node2ind: Dict[GraphNode, int] = {
            v: i for i, v in enumerate(ordered_nodes)
        }
        id2node: Dict[str, GraphNode] = {node.id: node for node in node2ind.keys()}
        adj_mx = dok_matrix((len(node2ind), len(node2ind)), dtype=np.short)
        for v, i in node2ind.items():
            for child in v.children:
                c_i = node2ind[child]
                try:
                    weight = max((w_req.volume for w_req in v.work_unit.worker_reqs), default=0.000001)
                except:
                    print('ln')
                adj_mx[i, c_i] = weight

        return ordered_nodes, adj_mx, id2node



from abc import ABC, abstractmethod
from typing import List, Callable, Optional

import numpy as np

from sampo.schemas.contractor import WorkerContractorPool
from sampo.schemas.resources import Worker
from sampo.schemas.time import Time


class ResourceOptimizer(ABC):

    @abstractmethod
    def optimize_resources(self,
                           worker_pool: WorkerContractorPool,
                           worker_team: List[Worker],
                           optimize_array: Optional[np.ndarray],
                           down_border: np.ndarray,
                           up_border: np.ndarray,
                           get_finish_time: Callable[[List[Worker]], Time]):
        """
        The resource optimization module. Optimizes `worker_team` using `get_finish_time` metric. Should optimize `worker_team` in-place.
        :param worker_pool: global resources pool
        :param worker_team: worker team to optimize
        :param optimize_array: a boolean array that says what positions should be optimized
        :param down_border: down border of optimization
        :param up_border: up border of optimization
        :param get_finish_time: optimization function that should give execution time based on worker team
        """
        ...



from typing import List, Optional, Type

from sampo.scheduler.base import SchedulerType
from sampo.scheduler.generic import GenericScheduler
from sampo.scheduler.heft.prioritization import prioritization
from sampo.scheduler.heft.time_computaion import calculate_working_time_cascade
from sampo.scheduler.resource.base import ResourceOptimizer
from sampo.scheduler.resource.coordinate_descent import CoordinateDescentResourceOptimizer
from sampo.scheduler.timeline.base import Timeline
from sampo.scheduler.timeline.just_in_time_timeline import JustInTimeTimeline
from sampo.scheduler.timeline.momentum_timeline import MomentumTimeline
from sampo.scheduler.utils.multi_contractor import get_worker_borders, run_contractor_search
from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.graph import GraphNode
from sampo.schemas.resources import Worker
from sampo.schemas.schedule_spec import WorkSpec
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.base_opt import dichotomy_int


class HEFTScheduler(GenericScheduler):

    def __init__(self,
                 scheduler_type: SchedulerType = SchedulerType.HEFTAddEnd,
                 resource_optimizer: ResourceOptimizer = CoordinateDescentResourceOptimizer(dichotomy_int),
                 timeline_type: Type = JustInTimeTimeline,
                 work_estimator: Optional[WorkTimeEstimator or None] = None):
        super().__init__(scheduler_type, resource_optimizer, timeline_type, prioritization,
                         self.get_default_res_opt_function(), work_estimator)
        self._timeline_type = timeline_type


class HEFTBetweenScheduler(HEFTScheduler):

    def __init__(self,
                 scheduler_type: SchedulerType = SchedulerType.HEFTAddBetween,
                 resource_optimizer: ResourceOptimizer = CoordinateDescentResourceOptimizer(dichotomy_int),
                 work_estimator: Optional[WorkTimeEstimator or None] = None):
        super().__init__(scheduler_type, resource_optimizer, MomentumTimeline, work_estimator)



from abc import ABC, abstractmethod

from sampo.scheduler.base import Scheduler
from sampo.scheduler.utils.local_optimization import OrderLocalOptimizer, ScheduleLocalOptimizer
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.schedule import Schedule
from sampo.schemas.time_estimator import WorkTimeEstimator


class InputPipeline(ABC):

    @abstractmethod
    def wg(self, wg: WorkGraph) -> 'InputPipeline':
        ...

    @abstractmethod
    def contractors(self, contractors: list[Contractor]) -> 'InputPipeline':
        ...

    @abstractmethod
    def work_estimator(self, work_estimator: WorkTimeEstimator) -> 'InputPipeline':
        ...

    @abstractmethod
    def node_order(self, node_order: list[GraphNode]) -> 'InputPipeline':
        ...

    @abstractmethod
    def optimize_local(self, optimizer: OrderLocalOptimizer, area: range) -> 'InputPipeline':
        ...

    @abstractmethod
    def schedule(self, scheduler: Scheduler) -> 'SchedulePipeline':
        ...


class SchedulePipeline(ABC):

    @abstractmethod
    def optimize_local(self, optimizer: ScheduleLocalOptimizer, area: range) -> 'SchedulePipeline':
        ...

    @abstractmethod
    def finish(self) -> Schedule:
        ...



class ApplyQueue:
    """
    Class represents the function apply sequence
    """

    def __init__(self):
        self.stack = []

    def add(self, f, args: tuple) -> 'ApplyQueue':
        """
        Adds function and args to queue. args should be (first arg, second arg, ...).
        The zero arg applied to `f` is the result of previous function in the queue.
        """
        self.stack.append((f, args))
        return self

    def apply(self, first_args):
        """
        Sequentially applies functions to the result of previous function and the args stored in queue.
        """
        res = first_args
        for f, args in self.stack:
            res = f(res, *args)
        return res

    def empty(self) -> bool:
        return len(self.stack) == 0
